/*
    data.c - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999, 2000  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"
#include "compat.h"
#include <stdlib.h>
#include <string.h>
#include "data.h"

static psiconv_color clone_color(psiconv_color color);
static psiconv_font clone_font(psiconv_font font);
static psiconv_border clone_border(psiconv_border border);
static psiconv_bullet clone_bullet(psiconv_bullet bullet);
static psiconv_all_tabs clone_all_tabs(psiconv_all_tabs all_tabs);
static void psiconv_free_style_aux(void *style);
static void psiconv_free_in_line_layout_aux(void * layout);
static void psiconv_free_paragraph_aux(void * paragraph);
static void psiconv_free_paint_data_section_aux(void * section);
static void psiconv_free_clipart_section_aux(void * section);

psiconv_character_layout psiconv_basic_character_layout(void)
{
  /* Make the structures static, to oblige IRIX */
  static struct psiconv_color_s black = 
  {
    0x00,                  /* red */
    0x00,                  /* green */
    0x00,                  /* blue */
  };
  static struct psiconv_color_s white = 
  {
    0xff,                  /* red */
    0xff,                  /* green */
    0xff,                  /* blue */
  };
  static struct psiconv_font_s font =
  {
    "Times New Roman",     /* name */
    3                      /* screenfont */
  };
  struct psiconv_character_layout_s cl = 
  {
    &black,                /* color               */
    &white,                /* back_color          */
    10.0,                  /* font_size           */
    psiconv_bool_false,    /* italic              */
    psiconv_bool_false,    /* bold                */
    psiconv_normalscript,  /* super_sub           */
    psiconv_bool_false,    /* underline           */
    psiconv_bool_false,    /* strikethrough       */
    &font,                  /* font                */
  };

  return psiconv_clone_character_layout(&cl);
}

psiconv_paragraph_layout psiconv_basic_paragraph_layout(void)
{
  static struct psiconv_font_s font =
  {
    "Times New Roman",     /* name */
    2                      /* screenfont */
  };
  static struct psiconv_color_s black = 
  {
    0x00,                  /* red */
    0x00,                  /* green */
    0x00,                  /* blue */
  };
  static struct psiconv_color_s white = 
  {
    0xff,                  /* red */
    0xff,                  /* green */
    0xff,                  /* blue */
  };
  static struct psiconv_border_s no_border =
  {
    psiconv_border_none,   /* kind                */
    1,                     /* thickness           */
    &black                 /* color               */
  };
  static struct psiconv_bullet_s bullet =
  {
    psiconv_bool_false,    /* on                  */
    10.0,                  /* font_size           */
    0x95,                  /* character           */
    psiconv_bool_true,     /* indent              */
    &black,                /* color               */
    &font,                 /* font                */
  };
  static struct psiconv_all_tabs_s tabs = 
  {
    0.64,                  /* normal              */
    NULL                   /* kind                */
  };
  struct psiconv_paragraph_layout_s pl =
  {
    &white,                /* back_color          */
    0.0,                   /* indent_left         */
    0.0,                   /* indent_right        */
    0.0,                   /* indent_first        */
    psiconv_justify_left,  /* justify_hor         */
    psiconv_justify_middle,/* justify_ver         */
    0.0,                   /* linespacing         */
    psiconv_bool_false,    /* linespacing_exact   */
    0.0,                   /* space_above         */
    0.0,                   /* space_below         */
    psiconv_bool_false,    /* keep_together       */
    psiconv_bool_false,    /* keep_with_next      */
    psiconv_bool_false,    /* on_next_page        */
    psiconv_bool_false,    /* no_widow_protection */
    0.0,                   /* left_margin         */
    &bullet,               /* bullet              */
    &no_border,            /* left_border         */
    &no_border,            /* right_border        */
    &no_border,            /* top_border          */
    &no_border,            /* bottom_border       */
    &tabs,                 /* tabs                */
  };
  psiconv_paragraph_layout res;
  
  if (!(pl.tabs->extras = psiconv_list_new(sizeof(struct psiconv_tab_s))))
    return NULL;
  res = psiconv_clone_paragraph_layout(&pl);
  psiconv_list_free(pl.tabs->extras);
  return res;
}
  
psiconv_color clone_color(psiconv_color color)
{
  psiconv_color result;
  if (!(result = malloc(sizeof(*result))))
    return NULL;
  *result = *color;
  return result;
}
  
psiconv_font clone_font(psiconv_font font)
{
  psiconv_font result;
  if(!(result = malloc(sizeof(*result))))
    goto ERROR1;
  *result = *font;
  if (!(result->name = strdup(result->name)))
    goto ERROR2;
  return result;
ERROR2:
  free(result);
ERROR1:
  return NULL;
}

psiconv_border clone_border(psiconv_border border)
{
  psiconv_border result;
  if (!(result = malloc(sizeof(*result))))
    goto ERROR1;
  *result = *border;
  if(!(result->color = clone_color(result->color)))
    goto ERROR2;
  return result;
ERROR2:
  free(result);
ERROR1:
  return NULL;
}

psiconv_bullet clone_bullet(psiconv_bullet bullet)
{
  psiconv_bullet result;
  if (!(result = malloc(sizeof(*result))))
    goto ERROR1;
  *result = *bullet;
  if (!(result->font = clone_font(result->font)))
    goto ERROR2;
  if (!(result->color = clone_color(result->color)))
    goto ERROR3;
  return result;
ERROR3:
  psiconv_free_font(result->font);
ERROR2:
  free(result);
ERROR1:
  return NULL;
}

psiconv_all_tabs clone_all_tabs(psiconv_all_tabs all_tabs)
{
  psiconv_all_tabs result;
  if (!(result = malloc(sizeof(*result))))
    goto ERROR1;
  *result = *all_tabs;
  if (!(result->extras = psiconv_list_clone(result->extras)))
    goto ERROR2;
  return result;
ERROR2:
  free(result);
ERROR1:
  return NULL;
}

psiconv_character_layout psiconv_clone_character_layout
                                           (psiconv_character_layout ls)
{
  psiconv_character_layout result;

  if (!(result = malloc(sizeof(*result))))
    goto ERROR1;
  *result = *ls;
  if (!(result->color = clone_color(result->color)))
    goto ERROR2;
  if (!(result->back_color = clone_color(result->back_color)))
    goto ERROR3;
  if (!(result->font = clone_font(result->font)))
    goto ERROR4;
  return result;
ERROR4:
  psiconv_free_color(result->back_color);
ERROR3:
  psiconv_free_color(result->color);
ERROR2:
  free(result);
ERROR1:
  return NULL;
}

psiconv_paragraph_layout psiconv_clone_paragraph_layout
                                          (psiconv_paragraph_layout ls)
{
  psiconv_paragraph_layout result;

  if (!(result = malloc(sizeof(*result))))
    goto ERROR1;
  *result = *ls;
  if (!(result->back_color = clone_color(result->back_color)))
    goto ERROR2;
  if (!(result->bullet = clone_bullet(result->bullet)))
    goto ERROR3;
  if (!(result->left_border = clone_border(result->left_border)))
    goto ERROR4;
  if (!(result->right_border = clone_border(result->right_border)))
    goto ERROR5;
  if (!(result->top_border = clone_border(result->top_border)))
    goto ERROR6;
  if (!(result->bottom_border = clone_border(result->bottom_border)))
    goto ERROR7;
  if (!(result->tabs = clone_all_tabs(result->tabs)))
    goto ERROR8;
  return result;
ERROR8:
  psiconv_free_border(result->bottom_border);
ERROR7:
  psiconv_free_border(result->top_border);
ERROR6:
  psiconv_free_border(result->right_border);
ERROR5:
  psiconv_free_border(result->left_border);
ERROR4:
  psiconv_free_bullet(result->bullet);
ERROR3:
  psiconv_free_color(result->back_color);
ERROR2:
  free(result);
ERROR1:
  return NULL;
}

psiconv_word_style psiconv_get_style (psiconv_word_styles_section ss, int nr)
{
  if (nr == 0)
    return ss->normal;
  else
    return psiconv_list_get(ss->styles,0xff - nr);
}

void psiconv_free_color (psiconv_color color)
{
  if (color)
    free(color);
}

void psiconv_free_border(psiconv_border border)
{
  if (border) {
    psiconv_free_color(border->color);
    free(border);
  }
}

void psiconv_free_font(psiconv_font font)
{
  if (font) {
    if (font->name) 
      free(font->name);
    free(font);
  }
}

void psiconv_free_bullet(psiconv_bullet bullet)
{
  if (bullet) {
    psiconv_free_color(bullet->color);
    psiconv_free_font(bullet->font);
    free(bullet);
  }
}

void psiconv_free_character_layout(psiconv_character_layout layout)
{
  if (layout) {
    psiconv_free_color(layout->color);
    psiconv_free_color(layout->back_color);
    psiconv_free_font(layout->font);
    free(layout);
  }
}

void psiconv_free_tab(psiconv_tab tab)
{
  if (tab)
    free(tab);
}

void psiconv_free_tabs(psiconv_all_tabs tabs)
{
  if (tabs) {
    psiconv_list_free(tabs->extras);
    free(tabs);
  }
}

void psiconv_free_paragraph_layout(psiconv_paragraph_layout layout)
{
  if (layout) {
    psiconv_free_color(layout->back_color);
    psiconv_free_bullet(layout->bullet);
    psiconv_free_border(layout->left_border);
    psiconv_free_border(layout->right_border);
    psiconv_free_border(layout->top_border);
    psiconv_free_border(layout->bottom_border);
    psiconv_free_tabs(layout->tabs);
    free(layout);
  }
}

void psiconv_free_style_aux(void *style)
{
  if(((psiconv_word_style) style)->name)
    free(((psiconv_word_style) style)->name);
  psiconv_free_character_layout(((psiconv_word_style) style)->character);
  psiconv_free_paragraph_layout(((psiconv_word_style) style)->paragraph);
}

void psiconv_free_word_style(psiconv_word_style style)
{
  if (style) {
    psiconv_free_style_aux(style);
    free(style);
  }
}

void psiconv_free_word_styles_section(psiconv_word_styles_section styles)
{
  if (styles) {
    psiconv_free_word_style(styles->normal);
    if (styles->styles) 
      psiconv_list_free_el(styles->styles,psiconv_free_style_aux);
  }
}

void psiconv_free_header_section(psiconv_header_section header)
{
  if (header)
    free(header);
}

void psiconv_free_section_table_entry(psiconv_section_table_entry entry)
{
  if (entry)
    free(entry);
}

void psiconv_free_section_table_section(psiconv_section_table_section section)
{
  if (section) 
    psiconv_list_free(section);
}

void psiconv_free_application_id_section(psiconv_application_id_section section)
{
  if (section) {
    if (section->name)
      free(section->name);
    free(section);
  }
}

void psiconv_free_in_line_layout_aux(void * layout)
{
  psiconv_free_character_layout(((psiconv_in_line_layout) layout)->layout);
}

void psiconv_free_in_line_layout(psiconv_in_line_layout layout)
{
  if (layout) {
    psiconv_free_in_line_layout_aux(layout);
    free(layout);
  }
}

void psiconv_free_in_line_layouts(psiconv_in_line_layouts layouts)
{
  if (layouts) 
    psiconv_list_free_el(layouts,&psiconv_free_in_line_layout_aux);
}

void psiconv_free_replacement(psiconv_replacement replacement)
{
  if (replacement) 
    free(replacement);
}

void psiconv_free_replacements(psiconv_replacements replacements)
{
  if (replacements) 
    psiconv_list_free(replacements);
}

void psiconv_free_paragraph_aux(void * paragraph)
{
  if(((psiconv_paragraph) paragraph)->text)
    free(((psiconv_paragraph) paragraph)->text);
  psiconv_free_character_layout(((psiconv_paragraph) paragraph)
                                                       ->base_character);
  psiconv_free_paragraph_layout(((psiconv_paragraph) paragraph)
                                                       ->base_paragraph);
  psiconv_free_in_line_layouts(((psiconv_paragraph) paragraph)
                                                       ->in_lines);
  psiconv_free_replacements(((psiconv_paragraph) paragraph)
                                                       ->replacements);
}

void psiconv_free_paragraph(psiconv_paragraph paragraph)
{
  if (paragraph) {
    psiconv_free_paragraph_aux(paragraph);
    free(paragraph);
  }
}

void psiconv_free_text_and_layout(psiconv_text_and_layout text)
{
  if (text)
    psiconv_list_free_el(text,&psiconv_free_paragraph_aux);
}

void psiconv_free_texted_section(psiconv_texted_section section)
{
  if (section) {
    psiconv_free_text_and_layout(section->paragraphs);
    free(section);
  }
}

void psiconv_free_page_header(psiconv_page_header header)
{
  if (header) {
    psiconv_free_character_layout(header->base_character_layout);
    psiconv_free_paragraph_layout(header->base_paragraph_layout);
    psiconv_free_texted_section(header->text);
    free(header);
  }
}

void psiconv_free_page_layout_section(psiconv_page_layout_section section)
{
  if (section) {
    psiconv_free_page_header(section->header);
    psiconv_free_page_header(section->footer);
    free(section);
  }
}

void psiconv_free_word_status_section(psiconv_word_status_section section)
{
  if (section) 
    free(section);
}

void psiconv_free_word_f(psiconv_word_f file)
{
  if (file) { 
    psiconv_free_page_layout_section(file->page_sec);
    psiconv_free_text_and_layout(file->paragraphs);
    psiconv_free_word_status_section(file->status_sec);
    psiconv_free_word_styles_section(file->styles_sec);
    free(file);
  }
}

void psiconv_free_texted_f(psiconv_texted_f file)
{
  if (file) {
    psiconv_free_page_layout_section(file->page_sec);
    psiconv_free_texted_section(file->texted_sec);
    free(file);
  }
}

void psiconv_free_paint_data_section_aux(void * section)
{
  if (((psiconv_paint_data_section) section)->red)
    free(((psiconv_paint_data_section)section) -> red);
  if (((psiconv_paint_data_section) section)->green)
    free(((psiconv_paint_data_section)section) -> green);
  if (((psiconv_paint_data_section) section)->blue)
    free(((psiconv_paint_data_section)section) -> blue);
}

void psiconv_free_paint_data_section(psiconv_paint_data_section section)
{
  if (section) {
    psiconv_free_paint_data_section_aux(section);
    free(section);
  }
}

void psiconv_free_pictures(psiconv_pictures section)
{
  if (section)
    psiconv_list_free_el(section,&psiconv_free_paint_data_section_aux);
}

void psiconv_free_jumptable_section (psiconv_jumptable_section section)
{
  if (section)
    psiconv_list_free(section);
} 

void psiconv_free_mbm_f(psiconv_mbm_f file)
{
  if (file) {
    psiconv_free_pictures(file->sections);
    free(file);
  }
}

void psiconv_free_sketch_section(psiconv_sketch_section sec)
{
  if (sec) {
    psiconv_free_paint_data_section(sec->picture);
    free(sec);
  }
}

void psiconv_free_sketch_f(psiconv_sketch_f file)
{
  if (file) {
    psiconv_free_sketch_section(file->sketch_sec);
    free(file);
  }
}

void psiconv_free_clipart_section_aux(void *section)
{
  if (section)
    free(((psiconv_clipart_section ) section)->picture);
}

void psiconv_free_clipart_section(psiconv_clipart_section section)
{
  if (section) {
    psiconv_free_clipart_section_aux(section);
    free(section);
  }
}

void psiconv_free_cliparts(psiconv_cliparts section)
{
  if (section)
    psiconv_list_free_el(section,&psiconv_free_clipart_section_aux);
}

void psiconv_free_clipart_f(psiconv_clipart_f file)
{
  if (file) {
    psiconv_free_cliparts(file->sections);
    free(file);
  }
}

void psiconv_free_file(psiconv_file file)
{
  if (file) {
    if (file->type == psiconv_word_file) 
      psiconv_free_word_f((psiconv_word_f) file->file);
    else if (file->type == psiconv_texted_file)
      psiconv_free_texted_f((psiconv_texted_f) file->file);
    else if (file->type == psiconv_mbm_file)
      psiconv_free_mbm_f((psiconv_mbm_f) file->file);
    else if (file->type == psiconv_sketch_file)
      psiconv_free_sketch_f((psiconv_sketch_f) file->file);
    else if (file->type == psiconv_clipart_file)
      psiconv_free_clipart_f((psiconv_clipart_f) file->file);
    free(file);
  }
}
